#include "display.h"

#include <mchd.h>

inline unsigned char Clamp(int value)
{
    if (value <= 0) {
        return 0;
    }

    if (value >= 255) {
        return 255;
    }

    return (unsigned char)value;
}


inline void ConvertYUVPixelToRGB(const double y, const double u, const double v, unsigned char &r, unsigned char &g, unsigned char &b)
{
    int R = (int)(y + 1.13983f * (v - 128.0f));
    int G = (int)(y - 0.39465f * (u - 128.0f) - 0.58060f * (v - 128.0f));
    int B = (int)(y + 2.03211f * (u - 128.0f));

    b = Clamp(B);
    g = Clamp(G);
    r = Clamp(R);
}


void ConvertBitmapFromYUYV(int width, int height, int pitch, unsigned char *bufferIn, unsigned char *bufferOut)
{
    unsigned char *rowIn = bufferIn;
    unsigned char *rowOut = bufferOut;

    double Y = 0;
    double U = 0;
    double V = 0;

    for (int y = 0; y < height; y++) {

        unsigned char *cursorIn = rowIn;
        unsigned char *cursorOut = rowOut;

        for (int x = 0; x < width; x++) {
            Y = (double)cursorIn[0];

            if (x % 2 == 0) {
                U = (double)cursorIn[1];
            } else {
                V = (double)cursorIn[1];
            }
            
            ConvertYUVPixelToRGB(Y, U, V, cursorOut[2], cursorOut[1], cursorOut[0]);

            cursorOut[3] = 0;

            cursorIn += 2;
            cursorOut += 4;
        }

        rowIn += pitch;
        rowOut += (width * 4);
    }
}


void ConvertBitmapFromYV12(int width, int height, const unsigned int pitch[3], unsigned char *const bufferIn[3], unsigned char *bufferOut)
{
    unsigned char *yRowIn = bufferIn[0];
    unsigned char *vRowIn = bufferIn[1];
    unsigned char *uRowIn = bufferIn[2];
    unsigned char *rowOut = bufferOut;

    double Y = 0;
    double U = 0;
    double V = 0;

    for (int y = 0; y < height; y++) {

        unsigned char *yCursorIn = yRowIn;
        unsigned char *vCursorIn = vRowIn;
        unsigned char *uCursorIn = uRowIn;
        unsigned char *cursorOut = rowOut;

        for (int x = 0; x < width; x++) {
            Y = (double)yCursorIn[0];

            if (x % 2 == 0) {
                U = (double)uCursorIn[0];
                V = (double)vCursorIn[0];
            }
            
            ConvertYUVPixelToRGB(Y, U, V, cursorOut[2], cursorOut[1], cursorOut[0]);

            cursorOut[3] = 0;

            yCursorIn++;

            if (x % 2 != 0) {
                uCursorIn++;
                vCursorIn++;
            }

            cursorOut += 4;
        }

        yRowIn += pitch[0];

        if (y % 2 != 0) {
            vRowIn += pitch[1];
            uRowIn += pitch[2];
        }

        rowOut += (width * 4);
    }
}


void DrawYUVSurface(HWND hwnd, const MCHD_VIDEO_SURFACE &surface, bool planar)
{
    if (hwnd == NULL) {
        return;
    }

    int width = surface.resolution.width;
    int height = surface.resolution.height;

    int rgb32BufferSize = surface.resolution.height * surface.resolution.width * 4;

    unsigned char *buffer = new unsigned char[rgb32BufferSize];

    if (!planar) {
        ConvertBitmapFromYUYV(width, height, surface.pitch[0], surface.data[0], buffer);
    } else {
        ConvertBitmapFromYV12(width, height, surface.pitch, surface.data, buffer);
    }

    HDC hDC = GetDC(hwnd);

    BITMAPINFOHEADER bih;
    bih.biSize = sizeof(BITMAPINFOHEADER);
    bih.biWidth = width;
    bih.biHeight = -(height);
    bih.biPlanes = 1;
    bih.biBitCount = 32;
    bih.biCompression = BI_RGB;
    bih.biSizeImage = 0;
    bih.biXPelsPerMeter = 0;
    bih.biYPelsPerMeter = 0;
    bih.biClrUsed = 0;
    bih.biClrImportant = 0;

    SetDIBitsToDevice(hDC, 
                      0, 
                      0, 
                      width, 
                      height, 
                      0, 
                      0, 
                      0, 
                      height,
                      (LPVOID)(buffer),
                      (BITMAPINFO*)&bih,
                      0);

    DeleteDC(hDC);

    delete[] buffer;
}


void DrawYUYVSurface(HWND hwnd, const MCHD_VIDEO_SURFACE &surface)
{
    DrawYUVSurface(hwnd, surface, false);
}


void DrawYV12Surface(HWND hwnd, const MCHD_VIDEO_SURFACE &surface)
{
    DrawYUVSurface(hwnd, surface, true);
}


LRESULT CALLBACK WndProc(HWND hwnd, UINT msg, WPARAM wParam, LPARAM lParam)
{
    switch(msg) {
      case WM_CLOSE:
        DestroyWindow(hwnd);
        break;

      case WM_DESTROY:
        PostQuitMessage(0);
        break;

      default:
        return DefWindowProc(hwnd, msg, wParam, lParam);
    }

    return 0;
}


HWND CreateWin32Window(int x, int y, int width, int height)
{
    WNDCLASSEX wc;
    HWND hwnd;

    HINSTANCE hInstance = (HINSTANCE)GetModuleHandle(NULL);

    wc.cbSize        = sizeof(WNDCLASSEX);
    wc.style         = 0;
    wc.lpfnWndProc   = WndProc;
    wc.cbClsExtra    = 0;
    wc.cbWndExtra    = 0;
    wc.hInstance     = hInstance;
    wc.hIcon         = LoadIcon(NULL, IDI_APPLICATION);
    wc.hCursor       = LoadCursor(NULL, IDC_ARROW);
    wc.hbrBackground = (HBRUSH)(COLOR_WINDOW+1);
    wc.lpszMenuName  = NULL;
    wc.lpszClassName = "mchdTestWindowClass";
    wc.hIconSm       = LoadIcon(NULL, IDI_APPLICATION);

    if (!RegisterClassEx(&wc)) {
        MessageBox(NULL, "Window Registration Failed!", "Error!", MB_ICONEXCLAMATION | MB_OK);
        return 0;
    }

    hwnd = CreateWindowEx(WS_EX_CLIENTEDGE,
                          "mchdTestWindowClass",
                          "MCHD Test Window",
                          WS_POPUP | WS_BORDER,
                          x, y, width, height,
                          NULL, NULL, hInstance, NULL);

    if (hwnd == NULL) {
        MessageBox(NULL, "Window Creation Failed!", "Error!", MB_ICONEXCLAMATION | MB_OK);
        return 0;
    }

    ShowWindow(hwnd, SW_SHOWNORMAL);
    UpdateWindow(hwnd);

    return hwnd;
}
